/*
 * npc.c - Nearly Perfect Crypto
 *
 * 21Apr2005 - Added more error checking, added code to prevent swapping
 *             to disk if root.
 * 03Apr2005 - Initial revision. Basic routines complete.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include "config.h"

typedef unsigned char BYTE;

#ifdef __FREEBSD__
#define LOCKING 0
#else
#define LOCKING 1
#endif

#define MAX_SIZE 2147483647

/*
 * actual memory setting
 */
void *guaranteed_memset(void *v,int c,size_t n)
{
        volatile char *p=v;
        while (n--) *p++=c;
        return v;
}

/*
 * take a file name and return a file handle
 */
FILE* get_file_handle(char * fname, int rw)
{
  FILE* fp;
  if(rw)
    fp = fopen(fname,"wb");
  else
    fp = fopen(fname,"rb");
  if(fp == NULL)
  {
    fprintf(stderr,"%s ERROR: cannot open file: %s\n",PACKAGE,fname);
    exit(-1);
  }
  return fp;
}

/*
 * get stats on a file
 */
struct stat get_stats(char *fname)
{
  struct stat statbuf;
  if(stat(fname,&statbuf) == -1)
  {
    fprintf(stderr,"%s ERROR: cannot find file: %s\n",PACKAGE,fname);
    exit(-1);
  }
  return(statbuf);
}

#ifndef WIN32
/*
 * If we are root (and not WIN32), lock memory to prevent paging to swap.
 */
void check_for_lockable_memory(int verbose)
{
  int lock = 0;
  if (getuid() == 0)
  {
    lock = mlockall(MCL_CURRENT|MCL_FUTURE);
    if (lock == -1)
    {
      fprintf(stderr,"%s: === Running as root but unable to lock memory to prevent paging to swap\n",PACKAGE);
      exit(-1);
    }
    lock = 1;
  }
  if (verbose)
  {
    if (lock)
      fprintf(stderr,"NOTE: Running as root and memory is locked from paging to disk\n");
    else
      fprintf(stderr,"WARNING: Running unprivileged and memory is not locked from paging to disk\n");
  }
}
#endif

/*
 * usage
 */
void usage(char *prog)
{
  fprintf(stderr,"USAGE: ");
  fprintf(stderr,"%s [opts] [-i <input file>] [-o output file] [-k keyfile]\n\n",prog);
  fprintf(stderr," opts are ? h O s v\n");
  fprintf(stderr,"  -? or -h    this help screen\n");
  fprintf(stderr,"  -O <offset> start reading from keyfile <offset> bytes in\n");
  fprintf(stderr,"  -s          track status of keyfile (can't use with -O)\n");
  fprintf(stderr,"  -v          verbose\n");
  fprintf(stderr,"  -V          print version and exit\n");
  fprintf(stderr,"\n");
}

int main(int argc, char **argv)
{
  FILE* ifp = NULL;
  FILE* ofp = NULL;
  FILE * kfp = NULL;
  char * ifile = NULL;
  char * ofile = NULL;
  char * kfile = NULL;
  FILE * sfile = NULL;
  BYTE iblock[4 * 4];
  BYTE oblock[4 * 4];
  BYTE kblock[4 * 4];
  int isize = 0;
  int ksize = 0;
  int osize = 0;
  int i;
  struct stat statbuf;
  unsigned long ifilesize = 0;
  unsigned long ofilesize = 0;
  unsigned long kfilesize = 0;
  unsigned long offset = 0;
  unsigned long stat_ptr = 0;
  int verbose=0, status=0;
  char *prog;
  char *status_file;
  char *t;
  extern char *optarg;
  extern int optind;
  extern int optopt;
  extern int opterr;
  char ch;
  
  prog = argv[0];

  while ((ch = getopt(argc, argv, "h?si:o:k:O:vV")) != EOF) 
    switch(ch)
    {
      case 'h':
      case '?':
        usage(prog);
        exit(0);
      case 's':
        if(status)
        {
          fprintf(stderr,"%s: === You cannot choose the status option more than once\n",PACKAGE);
          usage(prog);
          exit(-1);
        }
        status = 1;
        status_file = "~/.npc";
        sfile = get_file_handle(status_file,1);
        statbuf = get_stats(status_file);
        if(statbuf.st_size == 0)
          fprintf(stdout,"Creating new .npc file\n");
        else
        {
          i = fread(&stat_ptr, sizeof(unsigned long), 1, sfile);
          if(i <= 0)
          {
            fprintf(stderr,"%s: === Unable to properly read .npc file\n",PACKAGE);
            exit(-1);
          }
        }
        break;
      case 'i':
        if(ifile)
        {
          fprintf(stderr,"%s: You cannot use -i more than once\n",PACKAGE);
          exit(-1);
        }
        ifile = strdup(optarg);
        ifp = get_file_handle(ifile,0);
        statbuf = get_stats(ifile);
        ifilesize = (unsigned long)statbuf.st_size;
        if(ifilesize > MAX_SIZE)
        {
          fprintf(stderr,"%s: Error with %s, maximum file size is %d\n",ifile,PACKAGE,MAX_SIZE);
          exit(-1);
        }
        break;
      case 'o':
        if(ofile)
        {
          fprintf(stderr,"%s: You cannot use -o more than once\n",PACKAGE);
          exit(-1);
        }
        ofile = strdup(optarg);
        ofp = get_file_handle(ofile,1);
        break;
      case 'k':
        if(kfile)
        {
          fprintf(stderr,"%s: You cannot use -k more than once\n", PACKAGE);
          exit(-1);
        }
        kfile = strdup(optarg);
        kfp = get_file_handle(kfile,0);
        statbuf = get_stats(kfile);
        kfilesize = (unsigned long)statbuf.st_size;
        if(kfilesize > MAX_SIZE)
        {
          fprintf(stderr,"%s: Error with %s, maximum file size is %d\n",kfile,PACKAGE,MAX_SIZE);
          exit(-1);
        }
        break;
      case 'O':
        if(offset)
        {
          fprintf(stderr,"%s: === You cannot choose the offset option more than once\n",PACKAGE);
          usage(prog);
          exit(-1);
        }
        offset = strtoul(optarg,&t,10);
        if(*t == 0)
        {
          if(offset > MAX_SIZE)
          {
            fprintf(stderr,"%s: Error with offset %s, maximum offset is %d\n, \n",PACKAGE,optarg,MAX_SIZE);
            exit(-1);
          }
        }
        else
        {
          fprintf(stderr,"%s: Unable to convert requested offset: %s\n",PACKAGE,optarg);
          exit(-1);
        }
        break;
      case 'v':
        verbose = 1;
        break;
      case 'V':
        printf("%s %s\n", PACKAGE, VERSION);
        exit(0);
      default:
        usage(prog);
        exit(-1);
    }
  argc -= optind;
  argv += optind;

  /* post arg processing */
  if((!ifile) || (!ofile) || (!kfile))
  {
    fprintf(stderr,"%s: === You must supply an input file, a keyfile, and an output file\n",PACKAGE);
    exit(-1); 
  }
  if(ifilesize > kfilesize)
  {
    fprintf(stderr,"%s: === Given keyfile is smaller than input text\n",PACKAGE);
    exit(-1);
  }

  /* you can't do offset and status at the same time */
  if((status) && (offset))
  {
    fprintf(stderr,"%s: === You cannot use the offset feature and track status at the same time\n",PACKAGE);
  }

  if(status)
  {
    if(ifilesize > (kfilesize - stat_ptr))
    {
      fprintf(stderr,"%s: === Remaining keyspace in keyfile with status tracking is smaller than\n         input file\n",PACKAGE);
      exit(-1);
    }
  }

  if(offset)
  {
    if(ifilesize > (kfilesize - offset))
    {
      fprintf(stderr,"%s: === Remaining keyspace in keyfile with offset is smaller than input file\n",PACKAGE);
      exit(-1);
    }
  }

  /* end post arg processing */
  check_for_lockable_memory(verbose);

  ofilesize = ifilesize;
  /* if using status or offset, fseek the keyfile */
  if(status) fseek(kfp,stat_ptr,SEEK_SET); 
  if(offset) fseek(kfp,offset,SEEK_SET);

  /* main "crypto" loop */
  while(1)
  {
    guaranteed_memset(iblock, 0, 16);
    isize = fread(iblock, 1, 16, ifp);
    ksize = fread(kblock, 1, 16, kfp);
    if(isize <=0)
    {
      fprintf(stderr,"%s: === Unable to read data: %s\n",PACKAGE,ifile);
      exit(-1);
    }
    if(ksize <=0)
    {
      fprintf(stderr,"%s: === Unable to read data: %s\n",PACKAGE,kfile);
      exit(-1);
    }
    for(i = 0; i < isize; i++)
      oblock[i] = iblock[i] ^ kblock[i];
    osize = fwrite(oblock,isize,1,ofp);
    if(osize <= 0)
    {
      fprintf(stderr,"%s: === Unable to write data: %s\n",PACKAGE,ofile);
      exit(-1);
    }
    if(ofilesize<17) break;
    ofilesize -= isize;
  }
  if(status)
  {
    stat_ptr += kfilesize;
    if(fwrite(&stat_ptr,sizeof(unsigned long),1,sfile) != 1)
    {
      fprintf(stderr,"%s: === Error updating .npc file\n",PACKAGE);
      exit(-1);
    }
  }
  exit(0);
}
